import apache_beam as beam
import mysql.connector as mysql
from pydantic import BaseModel
from typing import Iterator, Dict, List


class MySQLConfig(BaseModel):
    host: str
    port: int = 3306
    username: str
    password: str
    database: str


class InvalidElementType(TypeError):
    """Error Class for when an Element is not the correct type"""


class ReadMySQL(beam.PTransform):
    """
    Query MySQL databases
    Args:
        dbconfig: Supports passing any config elements to the mysql.connector library. Most
        commonly this will be:
            host - host of database
            port - 3306 is the MySQL default
            username - User to connect as
            password -  Password for username
            database - The database to connect to
        query: The query to run against the database
    """
    def __init__(self, dbconfig: MySQLConfig, query: str):
        super().__init__()
        self.dbconfig = dbconfig
        self.query = query

    def expand(self, pcoll):
        return (
            pcoll
            | beam.Create([self.query])
            | beam.ParDo(_Query(self.dbconfig))
        )


class WriteToMySQL(beam.PTransform):
    """
    Insert Data into MySQL databases
    Args:
        dbconfig: Supports passing any config elements to the mysql.connector library. Most
        commonly this will be:
            host - host of database
            port - 3306 is the MySQL default
            username - User to connect as
            password -  Password for username
            database - The database to connect to
        table: The table name to insert into
        columns: The columns contained in the rows to be included
    """
    def __init__(self, dbconfig: MySQLConfig, table: str, columns: List):
        super().__init__()
        self.dbconfig = dbconfig
        self.table = table
        self.columns = columns

    def expand(self, pcoll):
        return (
            pcoll
            | beam.ParDo(_Insert(self.table, self.columns, self.dbconfig))
        )


class _Query(beam.DoFn):
    """An internal DoFn to be used in a PTransform. Not for external use.
    """
    def __init__(self, config: MySQLConfig):
        super().__init__(config)
        self.config = config
        self.conn = None
        self.cursor = None

    def start_bundle(self) -> None:
        self.conn = mysql.connect(**self.config.dict())
        self.cursor = self.conn.cursor(dictionary=True)

    def process(self, query) -> Iterator[Dict]:
        self.cursor.execute(query)
        for result in self.cursor:
            yield result

    def finish_bundle(self) -> None:
        if self.conn is not None:
            self.conn.close()


class _PutFn(beam.DoFn):
    """An internal DoFn to be used in a PTransform. Not for external use.
    """
    def __init__(self, table: str, columns: List, config: MySQLConfig):
        super().__init__()
        self.config = config
        self.table = table
        self.cols = columns
        self.conn = None
        self.cursor = None

    def start_bundle(self) -> None:
        self.conn = mysql.connect(**self.config.dict())
        self.cursor = self.conn.cursor(dictionary=True)

    def process(self, element) -> None:
        raise NotImplementedError

    def finish_bundle(self):
        self.conn.close()


class _Insert(_PutFn):
    """An internal DoFn to be used in a PTransform. Not for external use.
    """
    def process(self, element: List[Dict]) -> None:
        stmt = f"INSERT INTO `{self.table}` ({', '.join(self.cols)}) VALUES ({', '.join([f'%({col})s' for col in self.cols])});"
        if isinstance(element, list):
            self.cursor.executemany(stmt, element)
            self.conn.commit()
        else:
            raise InvalidElementType(f"Wrong element type. Expected List[Dict], received {type(element)}")



from beam_sink.sinks.mysql import ReadMySQL, WriteToMySQL, MySQLConfig
import apache_beam as beam
import unittest
import pytest
import json
from pydantic import ValidationError
import mysql.connector as mysql
import os

HOST = "localhost" if os.getenv("DEVELOP", False) else "mysql"


def setup_database(config: MySQLConfig):
    cnx = mysql.connect(host=config.host, port=config.port, username=config.username, password=config.password)
    cur = cnx.cursor()
    cur.execute("CREATE DATABASE IF NOT EXISTS thrillhouse")
    cnx.database = "thrillhouse"
    cur.execute("CREATE TABLE IF NOT EXISTS thrillhouse (id text, description text, amount text)")
    cnx.commit()
    cnx.close()


class TestMySQL(unittest.TestCase):
    def setUp(self) -> None:
        self.config = MySQLConfig(host=HOST, username="root", password="root", database="thrillhouse")
        setup_database(self.config)

    def test_mysql_query_returns(self):
        with beam.Pipeline() as p:
            p | 'ReadTable' >> ReadMySQL(self.config, "select * from thrillhouse")

    def test_success_of_assignment(self):
        config = {"host": HOST, "username": "root", "password": "root", "database": "thrillhouse"}
        dbconfig = MySQLConfig(**config)
        assert dbconfig.host == config["host"]
        assert dbconfig.username == config["username"]
        assert dbconfig.password == config["password"]
        assert dbconfig.database == config["database"]

    def test_validation_error(self):
        with pytest.raises(ValidationError):
            config = {"host": HOST, "username": "root", "database": "thrillhouse"}
            MySQLConfig(**config)

    def test_mysql_insert(self):
        with beam.Pipeline() as p:
            (
                p
                | 'ReadJson' >> beam.io.ReadFromText("tests/.data/test.jsonl")
                | 'Parse' >> beam.Map(lambda x: [json.loads(x)])
                | 'WriteData' >> WriteToMySQL(self.config, "thrillhouse", ["id", "description", "amount"])
            )

    def test_mysql_insert_with_partitions(self):
        with beam.Pipeline() as p:
            output = (
                p
                | 'ReadJson' >> beam.io.ReadFromText("tests/.data/test.jsonl")
                | 'Parse' >> beam.Map(lambda x: json.loads(x))
                | 'Batch' >> beam.BatchElements(2)
                | 'WriteData' >> WriteToMySQL(self.config, "thrillhouse", ["id", "description", "amount"])
            )




from setuptools import setup, find_packages


with open('requirements.txt') as f:
    requirements = f.read().splitlines()


setup_requirements = ['pytest-runner']

test_requirements = ['pytest']


setup(
    name='beam_sink',
    author='Mitchell Lisle',
    author_email='m.lisle90@gmail.com',
    description="An Apache Beam Sink Library for Databases and other Sinks.",
    install_requires=requirements,
    packages=find_packages(),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/mitchelllisle/beam-sink',
    version='0.12.0',
    zip_safe=False,
)
